/* Sigma - Simple greedy multiple alignment */
/* Copyright (C) 2007 Rahul Siddharthan     */
/* Licensed under the GNU GPL, version 2    */
/* See file COPYING in source distribution  */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "header.h"
#include "treer.h"
#include "commonroutines.h"
#include "setbgw.h"

extern int maxlen; 

char bases[4] = {'A', 'C', 'G', 'T'};



void printn(char *s, int n) 
{
    int m;
    
    for (m=0; m<n; m++)
        printf("%c",s[m]);
}



void printfrag(params *v, seq_frag *frag) 
{
    int n,m;
    GNode *anode=NULL, *subtree=NULL;

    getClosestAncestor(v,&anode,frag->seqindex,frag->nseqs);
    getSubTree(v,&subtree,frag->seqindex,frag->nseqs,&anode);
    
    printf("Number of sequences in frag: %d\n",frag->nseqs);
    printf("Tree: ");
    printTreeNewick(stdout,subtree);
    printf("\n");

    for (m=0; m<frag->nseqs; m++) {
        for (n=0; n<frag->seqlen; n++)
            printf("%10c",frag->seq[frag->seqindex[m]][n]);
        printf("\n");
    }
    for (n=0; n<frag->seqlen; n++)
        printf("%10.7f",frag->likelihood[n]);
    printf("\n\n");    
}



void init_sw_matrix(params *v, GPtrArray *seqs) 
{
    int n,maxlen=0;
    fastaseq *seq;
    
    for (n=0; n<seqs->len; n++) {
        seq= g_ptr_array_index(seqs,n);
        if (seq->seq->len > maxlen)
            maxlen=seq->seq->len;
    }
//    if (maxlen > v->maxfraglen)
//        maxlen = v->maxfraglen;
    swmat.fmat = malloc((maxlen+1)*sizeof(double));
    swmat.pmat = malloc((maxlen+1)*sizeof(double));
    swmat.lmat = malloc((maxlen+1)*sizeof(int));
}



void get_logodds_mean_variance(params *v, seq_frag *frag1, seq_frag *frag2, double *mean, double *variance, double *maxdiff) 
{
    // pick 100 random positions from frag1 and frag2, calculate mean
    // and variance.  crude is enough

    double logodds[1000];
    double probjoint,prob1,prob2, sqrtmp;
    int n, randn1,randn2;
    GNode *pNode = NULL;
    GNode *pSubTree3 = NULL;

    
    *mean = 0.0;
    *variance = 0.0;
    *maxdiff = 0.0;

    int *seqindex = (int*)calloc(sizeof(int), frag1->nseqs+frag2->nseqs);
    int m;

    for (m=0;m<frag1->nseqs;m++) {
        seqindex[m] = frag1->seqindex[m];
    }
    for (m=0;m<frag2->nseqs;m++) {
        seqindex[frag1->nseqs + m] = frag2->seqindex[m];
    } 

    getClosestAncestor(v, &pNode, seqindex, frag1->nseqs+frag2->nseqs);

    getSubTree(v, &pSubTree3, seqindex, frag1->nseqs+frag2->nseqs, &pNode);

    settransprobs(v, frag1, frag2, pSubTree3);
    
    /* One requires about 1000 values to be averaged to get a reasonably good (reproducible)
     * mean and variance */
    for (n=0; n<1000; n++) {
        randn1 = (int)((float)random()/((float) RAND_MAX) * frag1->seqlen);
        randn2 = (int)((float)random()/((float) RAND_MAX) * frag2->seqlen);

        prob1 = frag1->likelihood[randn1];
        prob2 = frag2->likelihood[randn2];
        probjoint = getTreeProb(v,pSubTree3,frag1,frag2,randn1,randn2);

        logodds[n] = gsl_sf_log(probjoint/prob1/prob2);
        *mean += logodds[n];
    }
    *mean /= 1000.0;
    for (n=0; n<1000; n++) {
        sqrtmp = (logodds[n]- *mean)*(logodds[n]- *mean);
        *variance += sqrtmp;
        if (sqrtmp > *maxdiff)
            *maxdiff = sqrtmp;
    }
    
    *variance = sqrt(*variance/1000.0);
    *maxdiff = sqrt(*maxdiff);
    free(seqindex);
    
}


double get_lpvalue_from_normal_dist(double logodds, double mean1, double variance1, double maxdiff1, int len) 
{
    double mean,variance;
    
    mean = mean1*((float)len);

    variance = variance1*sqrt((float)len);

    if (logodds < mean)
        return 0.0;
    else
        return -0.69314718055994529
            + gsl_sf_log_erfc((logodds-mean)/variance/1.4142135);

}


/* This returns the best gapless local match between the two frags, where
 * "best" is defined as best p-value, and the formula for the p-value is
 * the exact formula 1 - (1-p)^((sl1-l-1)(sl2-l-1)) derived in the former
 * sigma paper.  (Actually, the log of this is used.)
 *
 * Here p is the pvalue of two random sequences of length l having the
 * measured logodds or better.  Calculated in the get_lpvalue function
 * defined above.  sl1 and sl2 are the total sequence lengths of the two frags.
 *
 * For sufficiently small p, the approximation p(sl1-l-1)(sl2-l-1) is used.
 *
 * A matrix of f-values and a matrix of p-values is measured separately.  The
 * "corrected" p, i.e. with the length factors, is very large for short sequences,
 * so using it throughout will make it impossible to "nucleate" a sequence.
 * Therefore though the "corrected" p is stored in the p-matrix, the value of the
 * uncorrected p is used to decide whether to "terminate" the local alignment.
 */

void best_match_by_pvalue(params *v, seq_frag *frag1, seq_frag *frag2,
                          int *n1, int *n2, int *lbestout, double *bestout) 
{
    int sl1, sl2, bestm=0, bestn=0, lbest=0, m, n, laxis, ldiag, l_new,
        sl_min, m_init, n_init, ldiag_max, start1, start2;
    double best, f_old, f_new, p_old, p_new, logodds, pvalue,tempbestp,
        bestf=0.0;
    double prob1,prob2,prob3,worstp;
    double mean,variance,maxdiff;
    GNode *pSubTree3 = NULL;


    sl1 = frag1->seqlen;
    sl2 = frag2->seqlen;
    if ((sl1==0) || (sl2==0)) {
        *bestout = 0.0;
        *lbestout = 0;
        *n1 = 0;
        *n2 = 0;
        return;
    }

    worstp = 0.0;
    best = worstp;
    bestf = 0.0;


    /* The following code (before the for loop is to generate the subtrees */
    /* Create a new tree */

    int *seqindex = (int*)calloc(sizeof(int), frag1->nseqs+frag2->nseqs);

    for (m=0;m<frag1->nseqs;m++)
        seqindex[m] = frag1->seqindex[m];

    for (m=0;m<frag2->nseqs;m++){
        seqindex[frag1->nseqs + m] = frag2->seqindex[m];
    }

    GNode *pNode3;

    getClosestAncestor(v, &pNode3, seqindex, frag1->nseqs+frag2->nseqs);
    getSubTree(v, &pSubTree3, seqindex, frag1->nseqs+frag2->nseqs, &pNode3);

    free(seqindex);


    get_logodds_mean_variance(v, frag1, frag2, &mean, &variance,&maxdiff);
    settransprobs(v, frag1, frag2, pSubTree3);

    sl_min = (sl1 < sl2) ? sl1 : sl2;

    for (laxis = -sl2+1; laxis < sl1; laxis++) {
        if (laxis<0) {
            n_init = -laxis;
            m_init = 0;
            ldiag_max = ((sl2+laxis) < sl_min) ? (sl2+laxis+1) : sl_min+1;
        }
        else if (laxis > 0) {
            n_init = 0;
            m_init = laxis;
            ldiag_max = ((sl1-laxis) < sl_min) ? (sl1-laxis+1) : sl_min+1;
        }
        else { // laxis == 0
            n_init = 0;
            m_init = 0;
            ldiag_max = sl_min+1;
        }
        
        swmat.fmat[0]=0.0;
        swmat.pmat[0]=worstp;
        swmat.lmat[0]=0;

        for (ldiag=1; ldiag < ldiag_max; ldiag++) {
            m = m_init+ldiag;
            n = n_init+ldiag;
            
            prob1 = frag1->likelihood[m-1];
            prob2 = frag2->likelihood[n-1];
            
            prob3 = getTreeProb(v,pSubTree3,frag1,frag2,m-1,n-1);


            assert(prob3 > 0); assert (prob3 < 1.0);
            
            logodds = (double) gsl_sf_log(prob3/prob1/prob2);
            
            f_old = swmat.fmat[ldiag-1];
            p_old = swmat.pmat[ldiag-1];
            l_new = swmat.lmat[ldiag-1] + 1;

            f_new = f_old + logodds;
            pvalue = get_lpvalue_from_normal_dist(f_new,mean,variance,maxdiff,l_new);

            // FIXME: this value is arbitrary... should there be q-dependence?
            if (pvalue < gsl_sf_log(0.2)) {
                if (pvalue < -6) // use approximation to avoid underflows
                    p_new = log((sl1-l_new+1.0)*(sl2-l_new+1.0)) + pvalue;
                else // exact expression
                    p_new = log(1.0-exp((sl1-l_new+1.0)*(sl2-l_new+1.0)*gsl_sf_log(1.0-exp(pvalue))));
            }
            else {
                p_new = worstp;
                f_new = 0.0;
                l_new = 0;
            }
                
                
            swmat.fmat[ldiag] = f_new;
            swmat.pmat[ldiag] = p_new;
            swmat.lmat[ldiag] = l_new;
            
            if (p_new < best) {
                best = p_new;
                bestf = f_new;
                lbest = l_new;
                bestm = m;
                bestn = n;
            }
        }
    }
    

    if ((best < log(v->pmatch))&& (lbest >= 5)) {
        start1 = bestm-lbest;
        start2 = bestn-lbest;

        tempbestp = best;
        bestn = 0;
        
        *lbestout = lbest;
        *bestout = best;

        /* fixup for poor beginnings: sometimes an initial
         * misalignment continues to a good alignment, so find the
         * starting position that gives truly the best pvalue
         */
        swmat.fmat[0] = 0;
        for (n=1; n<=lbest; n++) {
            /* fill up fmat with the appropriate values; */
            prob1 = frag1->likelihood[start1+n-1];
            prob2 = frag2->likelihood[start2+n-1];
            prob3 = getTreeProb(v,pSubTree3,frag1,frag2,start1+n-1,start2+n-1);
            assert(prob3 > 0); assert (prob3 < 1.0);
            logodds = (double) gsl_sf_log(prob3/prob1/prob2);
            swmat.fmat[n] = swmat.fmat[n-1] + logodds;
        }
        /* consistency check */
        assert(fabs((bestf-swmat.fmat[lbest])/bestf) < 0.01);
        bestf = swmat.fmat[lbest];
        
        for (n=0; n<=lbest; n++) {
            pvalue = get_lpvalue_from_normal_dist(bestf-swmat.fmat[n],mean,variance,maxdiff,lbest-n);
            if (pvalue==0)
                continue;
            if (pvalue < -6) // use approximation to avoid underflows
                pvalue = log((sl1-(lbest-n)+1.0)*(sl2-(lbest-n)+1.0)) + pvalue;
            else // exact expression
                pvalue = log(1.0-exp((sl1-(lbest-n)+1.0)*(sl2-(lbest-n)+1.0)*gsl_sf_log(1.0-exp(pvalue))));
            if (pvalue < tempbestp) {
                tempbestp = pvalue;
                bestn = n;
            }
        }
        *n1 = start1+bestn;
        *n2 = start2+bestn;
        *lbestout = lbest-bestn;
        *bestout = tempbestp;
    }
    else {
        *n1 = 0;
        *n2 = 0;
        *lbestout = 0;
        *bestout = worstp;
    }
}



void replace_string(char **s1, char **s2, char c) 
{
    int n;

    if (*s1)
        free(*s1);
    n=strlen(*s2);
    if (c=='\0')
        *s1 = calloc(n+1,sizeof(char));
    else
        *s1 = calloc(n+2,sizeof(char));
    strncpy(*s1, *s2, n+1);
    if (c=='\0')
        (*s1)[n]='\0';
    else {
        (*s1)[n]=c;
        (*s1)[n+1]='\0';
    }
}


void reset_swept(params *v) /* avoid sweeping twice from same frag */
{
    seq_frag *frag;
    int n;

    for (n=0; n < v->seqfraglist->len; n++) {
        frag = g_ptr_array_index(v->seqfraglist,n);
        frag->swept = 0;
    }
}


void sweep_limits(char **lim, seq_frag *frag, int ns, int dir) 
{
    int n;

    if ((frag==NULL)||(frag->swept))
        return;

    frag->swept = 1;
    if (dir) { /* left sweep, right lims */
        if (strcmp(frag->limits_r[ns],*lim) > 0) {
            replace_string(&(frag->limits_r[ns]),lim,'\0');
        }
        for (n=0; n<frag->nseqs; n++) 
            sweep_limits(lim,frag->neighbours_l[frag->seqindex[n]],ns,dir);
    }
    else { /* right sweep, left lims */
        if (strcmp(frag->limits_l[ns],*lim) < 0) {
            replace_string(&(frag->limits_l[ns]),lim,'\0');
        }
        for (n=0; n<frag->nseqs; n++)
            sweep_limits(lim,frag->neighbours_r[frag->seqindex[n]],ns,dir);
    }
    
}

            
        



void update_limits(params *v, seq_frag *frag) 
{
    int m,ns,ns2,is_diag;
    seq_frag *fragn;

    for (ns=0; ns <v->nseqs; ns++) {
        is_diag = 0;
        for (m=0; m < frag->nseqs; m++)
            if (frag->seqindex[m]==ns) {
                is_diag = 1;
                break;
            }

        if (is_diag) {
        /* diagonal limits */
            replace_string(&((frag->limits_l)[ns]),&((frag->seqlabel)[ns]),'\0');
            replace_string(&((frag->limits_r)[ns]),&((frag->seqlabel)[ns]),'\0');
        }
        else  /* off-diag */
            /* replace limit for ns with most stringent limit of a
               neighbouring fragment */
            for (m=0; m < frag->nseqs; m++) {
                ns2 = frag->seqindex[m];
                fragn = frag->neighbours_l[ns2];
                if (fragn) {
                    if (strcmp(fragn->limits_l[ns],frag->limits_l[ns])>0)
                        replace_string(&(frag->limits_l[ns]),&(fragn->limits_l[ns]),'\0');
                }
                fragn = frag->neighbours_r[ns2];
                if (fragn) {
                    if (strcmp(fragn->limits_r[ns],frag->limits_r[ns])<0) 
                        replace_string(&(frag->limits_r[ns]),&(fragn->limits_r[ns]),'\0');
                }
            }
    }

    for (ns = 0; ns < v->nseqs; ns++) {
        reset_swept(v);
        sweep_limits(&(frag->limits_l[ns]),frag,ns,0);
        reset_swept(v);
        sweep_limits(&(frag->limits_r[ns]),frag,ns,1);
    }
}

    

int inconsistent_frags(seq_frag *frag1, seq_frag *frag2)
{
    int n1, n2, s1, s2;
    char *label, *lim_l, *lim_r;
    
    for (n1=0; n1< frag1->nseqs; n1++) {
        s1 = frag1->seqindex[n1];
        for (n2=0; n2 < frag2->nseqs; n2++) {
            s2 = frag2->seqindex[n2];
            if (s2 == s1)
                return 1;
            else {
                label = frag2->seqlabel[s2];
                lim_l = frag1->limits_l[s2];
                lim_r = frag1->limits_r[s2];
                if ((strcmp(label,lim_l)<=0)||(strcmp(label,lim_r)>=0))
                    return 1;
                label = frag1->seqlabel[s1];
                lim_l = frag2->limits_l[s1];
                lim_r = frag2->limits_r[s1];
                if ((strcmp(label,lim_l)<=0)||(strcmp(label,lim_r)>=0))
                    return 1;
            }
        }
    }
    return 0;
}


gint compare_matches(gconstpointer *a, gconstpointer *b) 
{
    match *m1, *m2;

    m1 = (match*)(*a);
    m2 = (match*)(*b);
    
    if (m1->p < m2->p)
        return -1;
    else if (m1->p > m2->p)
        return 1;
    else
        return 0;
}


void make_match_list(params *v, GPtrArray **matchlist) 
{
    int n1, n2;
    seq_frag *frag1, *frag2;
    match *onematch;
    
    if (*matchlist) {
        for (n1=0; n1<(*matchlist)->len; n1++) {
            onematch = g_ptr_array_index(*matchlist,n1);
            free(onematch);
        }
        g_ptr_array_free(*matchlist, TRUE);
    }
    
    *matchlist = g_ptr_array_new();
    onematch = calloc(sizeof(match),1);
    for (n1=0; n1<v->seqfraglist->len - 1; n1++) {
        frag1 = g_ptr_array_index(v->seqfraglist,n1);
        for (n2 = n1+1; n2 < v->seqfraglist->len; n2++) {
            frag2 = g_ptr_array_index(v->seqfraglist,n2);
            if (inconsistent_frags(frag1,frag2))
                continue;
			
            onematch->frag1 = frag1;
            onematch->frag2 = frag2;

            best_match_by_pvalue(v,frag1,frag2,&(onematch->n1),&(onematch->n2),
                                 &(onematch->l),&(onematch->p));

            if (onematch->l > 0) {
                g_ptr_array_add(*matchlist,onematch);
                onematch = calloc(sizeof(match),1);
            }
        }
    }
    free(onematch);
    g_ptr_array_sort((*matchlist), (GCompareFunc) compare_matches);
}


void fuse_one_match(params *v, match *onematch) 
{
    int n,ns;
    seq_frag *f1, *f2, *f3, *f4, *f5;
    
    f1 = onematch->frag1;
    f2 = onematch->frag2;

    if (inconsistent_frags(f1,f2))
        return;

    f3 = (seq_frag*) new_seqfrag(v);
    f4 = (seq_frag*) new_seqfrag(v);
    f5 = (seq_frag*) new_seqfrag(v);

 

    /* set seqs of various frags */
    f3->nseqs = f1->nseqs + f2->nseqs;
    f4->nseqs = f1->nseqs;
    f5->nseqs = f2->nseqs;

    f3->seqlen = onematch->l;
    f4->seqlen = f1->seqlen - onematch->n1 - onematch->l;
    f5->seqlen = f2->seqlen - onematch->n2 - onematch->l;
    
    f1->seqlen = onematch->n1;
    f2->seqlen = onematch->n2;


    for (n=0; n<f1->nseqs; n++) {
        ns = f1->seqindex[n];
        f3->seqindex[n] = ns;
        f4->seqindex[n] = ns;

        (f3->seq)[ns] = (char*)((f1->seq)[ns] + onematch->n1);
        (f4->seq)[ns] = (char*)((f1->seq)[ns] + onematch->n1 + onematch->l);
        (f3->transprob)[ns] = (double*)((f1->transprob)[ns] + onematch->n1*4);
        (f4->transprob)[ns] = (double*)((f1->transprob)[ns] + (onematch->n1 + onematch->l)*4);
        

        /* relabel */
        replace_string(&(f3->seqlabel[ns]),&(f1->seqlabel[ns]),'1');
        replace_string(&(f4->seqlabel[ns]),&(f1->seqlabel[ns]),'2');
        /* trickery to relabel first frag */
        replace_string(&(f1->seqlabel[ns]),&(f4->seqlabel[ns]),'\0');
        f1->seqlabel[ns][strlen(f1->seqlabel[ns])-1] = '0';
        /* neighbours */
        f4->neighbours_r[ns] = f1->neighbours_r[ns];
        f1->neighbours_r[ns] = f3;
        f3->neighbours_l[ns] = f1;
        f3->neighbours_r[ns] = f4;
        f4->neighbours_l[ns] = f3;
    }
    for (n=0; n<f2->nseqs;n++) {
        ns = f2->seqindex[n];
        f3->seqindex[n+f1->nseqs] = ns;
        f5->seqindex[n] = ns;

        (f3->seq)[ns] = (char*)((f2->seq)[ns] + onematch->n2);
        (f5->seq)[ns] = (char*)((f2->seq)[ns] + onematch->n2 + onematch->l);
        (f3->transprob)[ns] = (double*)((f2->transprob)[ns] + onematch->n2*4);
        (f5->transprob)[ns] = (double*)((f2->transprob)[ns] + (onematch->n2 + onematch->l)*4);

        /* relabel */
        replace_string(&(f3->seqlabel[ns]),&(f2->seqlabel[ns]),'1');
        replace_string(&(f5->seqlabel[ns]),&(f2->seqlabel[ns]),'2');
        /* trickery */
        replace_string(&(f2->seqlabel[ns]),&(f5->seqlabel[ns]),'\0');
        f2->seqlabel[ns][strlen(f2->seqlabel[ns])-1] = '0';
        /* neighbours */
        f5->neighbours_r[ns] = f2->neighbours_r[ns];
        f2->neighbours_r[ns] = f3;
        f3->neighbours_l[ns] = f2;
        f3->neighbours_r[ns] = f5;
        f5->neighbours_l[ns] = f3;
    }

    /* likelihoods */

    f3->likelihood = calloc(f3->seqlen,sizeof(double));
    
    set_likelihoods(v,&f3);
    

#ifdef TEST_LIKELIHOOD    
    if (f3->nseqs >= 3) 
        printfrag(v,f3);
#endif    
    
    f4->likelihood = (double *)(f1->likelihood + onematch->n1 + onematch->l);
    f5->likelihood = (double *)(f2->likelihood + onematch->n2 + onematch->l);
    
    /* limits */
    for (n=0; n<v->nseqs; n++) {
        replace_string(&(f4->limits_l[n]),&(f1->limits_l[n]),'\0');
        replace_string(&(f5->limits_l[n]),&(f2->limits_l[n]),'\0');
        
        replace_string(&(f4->limits_r[n]),&(f1->limits_r[n]),'\0');
        replace_string(&(f5->limits_r[n]),&(f2->limits_r[n]),'\0');        
    }

    g_ptr_array_add(v->seqfraglist,f3);
    g_ptr_array_add(v->seqfraglist,f4);
    g_ptr_array_add(v->seqfraglist,f5);
    update_limits(v,f3);

}


int seqindices_match(seq_frag *frag1, seq_frag *frag2) 
{
    int n1,n2, is_in;
    if (frag1->nseqs != frag2->nseqs)
        return 0;
    
    for (n1=0; n1<frag1->nseqs; n1++) {
        is_in = 0;
        for (n2=0; n2<frag2->nseqs; n2++) 
            if (frag1->seqindex[n1]==frag2->seqindex[n2]) {
                is_in=1;
                break;
            }
        if (is_in==0)
            return 0;
    }
    return 1;
}

            


void expand_match(match *onematch, GPtrArray **outmatches) 
{
    /* In case of earlier fusions, a match may now be split over
       multiple fragments; recover from this situation */

    match *tmpmatch, *nextmatch;
    seq_frag *nextfrag1, *nextfrag2;
    int  s1,s2, n1, n2;
    
    *outmatches = g_ptr_array_new();
    tmpmatch=(match*)calloc(1, sizeof(match));
    tmpmatch->frag1 = onematch->frag1;
    tmpmatch->frag2 = onematch->frag2;
    tmpmatch->n1 = onematch->n1;
    tmpmatch->n2 = onematch->n2;
    tmpmatch->l = onematch->l;
    tmpmatch->p = onematch->p;

    
    s1 = tmpmatch->frag1->seqindex[0];
    s2 = tmpmatch->frag2->seqindex[0];
    while ((tmpmatch->n1 + tmpmatch->l > tmpmatch->frag1->seqlen)||
           (tmpmatch->n2 + tmpmatch->l > tmpmatch->frag2->seqlen)) {
        n1=0;
        while ((tmpmatch->frag1->seqindex[n1] != s1)
               && (n1 < tmpmatch->frag1->nseqs))
            n1++;
        assert(n1<tmpmatch->frag1->nseqs);
        nextfrag1 = tmpmatch->frag1->neighbours_r[s1];

        n2=0;
        while ((tmpmatch->frag2->seqindex[n2] != s2)
               && (n2 < tmpmatch->frag2->nseqs))
            n2++;
        assert(n2 < tmpmatch->frag2->nseqs);
        nextfrag2 = tmpmatch->frag2->neighbours_r[s2];
		if (nextfrag2==NULL) return;
        if ((tmpmatch->frag1->seqlen <= tmpmatch->n1)
            &&(tmpmatch->frag2->seqlen <= tmpmatch->n2)) {
            /* match lies outside both current frags */
            assert(nextfrag1 != NULL);
            assert(nextfrag2 != NULL);
            tmpmatch->n1 -= tmpmatch->frag1->seqlen;
            tmpmatch->n2 -= tmpmatch->frag2->seqlen;
            tmpmatch->frag1 = nextfrag1;
            tmpmatch->frag2 = nextfrag2;
        }
        else if (tmpmatch->frag1->seqlen <= tmpmatch->n1) {
            /* only frag1 is outside bounds */
            assert(nextfrag1 != NULL);
            tmpmatch->n1 -= tmpmatch->frag1->seqlen;
            tmpmatch->frag1 = nextfrag1;
        }
        else if (tmpmatch->frag2->seqlen <= tmpmatch->n2) {
            assert(nextfrag2!=NULL);
            tmpmatch->n2 -= tmpmatch->frag2->seqlen;
            tmpmatch->frag2 = nextfrag2;
        }
        else { /* Both frags contain some alignable sequence;
                  we align what exists and continue */
            nextmatch = malloc(sizeof(match));
            nextmatch->frag1 = tmpmatch->frag1;
            nextmatch->frag2 = tmpmatch->frag2;
            nextmatch->p = tmpmatch->p;
            
            if ((tmpmatch->frag1->seqlen - tmpmatch->n1)
                < (tmpmatch->frag2->seqlen - tmpmatch->n2)) {
                /* frag1 contains less alignable sequence than frag2 */
                nextmatch->frag1 = nextfrag1;
                nextmatch->n1 = 0;
                nextmatch->l = tmpmatch->n1 + tmpmatch->l
                    - tmpmatch->frag1->seqlen;
                nextmatch->n2 = tmpmatch->n2 + tmpmatch->l - nextmatch->l;
                tmpmatch->l = tmpmatch->frag1->seqlen - tmpmatch->n1;
            }
            else { /* frag2 contains less alignable seq than frag1 */
                nextmatch->frag2 = nextfrag2;
                nextmatch->n2 = 0;
                nextmatch->l = tmpmatch->n2 + tmpmatch->l
                    - tmpmatch->frag2->seqlen;
                nextmatch->n1 = tmpmatch->n1 + tmpmatch->l - nextmatch->l;
                tmpmatch->l = tmpmatch->frag2->seqlen - tmpmatch->n2;
            }
            g_ptr_array_add(*outmatches,tmpmatch);
            tmpmatch=nextmatch;
        }
    }
    g_ptr_array_add(*outmatches,tmpmatch);
    return;
}




int fuse_match_list(params *v, GPtrArray *matchlist) 
{
    int m,n,aligned=0;
    match *onematch, *tmpmatch;
    GPtrArray *outmatches;
//    printf("\n ***************");
    /* assume matchlist is sorted, carry out fusions */
    for (m=0; m<matchlist->len; m++) {
        onematch = g_ptr_array_index(matchlist,m);
//        printf("fusing fvalue %f\n",onematch->p);

        expand_match(onematch, &outmatches);
        if (outmatches) {
            for (n=0; n<outmatches->len; n++) {
                tmpmatch=g_ptr_array_index(outmatches,n);
                if (inconsistent_frags(tmpmatch->frag1,tmpmatch->frag2))
                    break;
                fuse_one_match(v,tmpmatch);
                aligned=1;
            }
            for (n=0; n<outmatches->len; n++) {
                tmpmatch = g_ptr_array_index(outmatches,n);
                cfree(tmpmatch);
            }
            g_ptr_array_free(outmatches,TRUE);
        }
    }
    return aligned;
}


#if 0

void get_sandwiched_frags(params *v, seq_frag *frag1, seq_frag *frag2,
                          GPtrArray **fraglist) 
{
    /* get list of frags that exist between frag1 and frag2 */
}


void align_sandwiched_frags(params *v, GPtrArray *fraglist) 
{
    /* repeatedly find longest gapless maximal matches between pairs
     * of frags, and align, till none left */
}


void filter_small_overhangs(params *v) 
{
    /* in cases where there are short (<10bp) unaligned regions in
     * between larger aligned blocks, find best GAPLESS MAXIMAL
     * alignments and align them
     *
     * strategy: put all unaligned regions between these two frags
     * in a new fraglist, repeatedly find and perform best match until
     * there are no remaining matches.  Best match here is not by pvalue
     * but purely by length, and within length, by number of matches --
     * implement separately?  Fuse_one_match can be reused.
     */

    aligned = 1;
    while (aligned) {
        if (short_segs_exist(v,&frag1,&frag2)) {
            /* frag1 and frag2 sandwich shorter unaligned
             * fragments: align them */
            get_sandwiched_frags(v,frag1,frag2,&fraglist);
            align_sandwiched_frags(v,fraglist);
        }
        else
            aligned=0;
    }
    
}

#endif

void blockseq(params *v)
{
    int aligned;
    GPtrArray *matchlist;

    matchlist = NULL;

    do {
        aligned=0;
        make_match_list(v,&matchlist);
        if (matchlist->len > 0) {
            aligned=fuse_match_list(v,matchlist);
        }
    } while (aligned);
}
