/* Sigma - Simple greedy multiple alignment */
/* Copyright (C) 2007 Rahul Siddharthan     */
/* Licensed under the GNU GPL, version 2    */
/* See file COPYING in source distribution  */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "sigma.h"
#include "fasta.h"
#include "commonroutines.h"
#include "treer.h"


double condbgprob(params *v, int m, char neigh_l, char neigh_r) 
{
    /* returns the conditional background probability given
     * two neighbours */
    /* if both neighbours are present, return
     * P(m|l,r) = P(l,r|m) P(m) / normalisation
     * where P(l,r|m) = P(l|m)P(r|m)
     */
    int nl,nr,m1;
    double p_m,tot;

    if (v->correlbg==0)
        return 0.25;
    if (v->correlbg==1)
        return v->singlesite[m];
    
    nl = base2num(neigh_l);
    nr = base2num(neigh_r);

    if ((nl== -1) && (nr== -1))
        return v->singlesite[m];

    if (nl == -1)
        return v->dinucleotide_r[m*4+nr];
    else if (nr == -1)
        return v->dinucleotide_l[m*4+nl];
    else {
        p_m = v->dinucleotide_l[m*4+nl]*v->dinucleotide_r[m*4+nr]*v->singlesite[m];
        tot = 0.0;
        for (m1=0; m1<4; m1++) 
            tot += v->dinucleotide_l[m1*4+nl]*v->dinucleotide_r[m1*4+nr]*v->singlesite[m1];
        return p_m/tot;
    }
}



void mutprob(params *v, char neigh_l, char neigh_r, double *mutmat) 
{
    /* return the transition matrix P(alpha_n|anc, neighbours) according to
     * P(alpha_n|anc, alpha_{n-1}, alpha_{n+1})
     *   = P(anc | alpha_n) P(alpha_n | alpha_{n-1}, alpha_{n+1})/(norm. over alpha_n)
     * P(alpha_n | n-1, n+1) is not a simple product.. Bayesian estimate?
     * or is it better to use one side, eg n-1, only?
     * 
     * If v->usetransitions==0, all anc transitions equally likely (0.3 each),
     * diagonal elements are zero.
     * else use appropriate value
     */

    /* assume mutmat is already allocated */
    /* calling convention: mutmat[n*4+m] -> P(n|m), i.e. n = anc */

    int m,n;
    double tot,panc;
    
    for (n=0; n<4; n++) {
        tot = 0.0;
        for (m=0; m<4; m++) {
            if (v->usetransitions)
                panc = v->invtransitionmat[n*4+m];
            else {
                if (m==n)
                    panc = 0.0;
                else
                    panc = 1.0/3.0;
            }
            mutmat[n*4+m] = panc * condbgprob(v,m,neigh_l,neigh_r);
            tot += mutmat[n*4+m];
        }
        for (m=0; m<4; m++)
            mutmat[n*4+m] /= tot;
    }
}



void transitionprob(double *mutprob, double q, double *transprob) 
{
    /* return q^(I-P) where P = mutprob matrix) */
    /* transprob takes the place of the bgw in the frag; and it is updated
     * for each call of best_match(), for both fragments being aligned
     */
    /* upfront expense, but the alignment itself should go much faster
     * since one only needs to read off the values...
     */

    int n,m,p,k;
    double *cumulmat, *tmpmat, mu, cumulmu, tot;
    
    cumulmat = malloc(16*sizeof(double));
    tmpmat = malloc(16*sizeof(double));
        
    mu = -log(q);
    
    for (n=0; n<4; n++)
        for (m=0;m<4; m++)
            cumulmat[n*4+m] = mutprob[n*4+m];

    cumulmu = mu;
    
    for (m=0; m<4; m++)
        for (n=0; n<4; n++)
            transprob[n*4+m] = (m==n) ? 1.0 : 0.0;
    
    k=1;
    while (fabs(cumulmu) > 0.001) {
        // transprob += cumulmu*cumulmat;
        for (n=0; n<4; n++)
            for (m=0; m<4; m++)
                transprob[n*4+m] += cumulmu * cumulmat[n*4+m];
        k += 1;
        cumulmu *= mu/k;
        // cumulmat *= mutprob;
        for (n=0; n<4; n++)
            for (m=0; m<4; m++) {
                tmpmat[n*4+m] = 0.0;
                for (p=0; p<4; p++)
                    tmpmat[n*4+m] += cumulmat[n*4+p]*mutprob[p*4+m];
            }
        for (n=0; n<4; n++)
            for (m=0; m<4; m++) 
                cumulmat[n*4+m] = tmpmat[n*4+m];
    
    }
    for (m=0; m<4; m++) {
        tot = 0.0;
        for (n=0; n<4; n++) 
            tot += transprob[n*4+m];
        for (n=0; n<4; n++)         
            transprob[n*4+m] /= tot;
    }
    
    
    free(cumulmat);
    free(tmpmat);
}



    
void settransprobs(params *v, seq_frag *frag1, seq_frag *frag2, GNode *tree) 
{
    int n,ns,m,anc;
    double mutmat[16], transmat[16], q;
    GNode *child;
    seq_frag *frag;

    /* traverse the tree; if node, calculate matrix and insert into
     * the data part of the node; if leaf, set the transition probs for
     * the frag corresponding to that leaf. */
    
    /* if a leaf base is not ACGT, set transprob as 0.25 */

    if (tree->children) { /* node */
        /* insert transition probs into this node, if it is not the root */
        if (tree->parent) {
            q = ((treest*)(tree->data))->depth;
            mutprob(v,'\0','\0',mutmat);
            transitionprob(mutmat,q,((treest*)(tree->data))->transmat);
        }
        
        child = tree->children;
        while (child) {
            settransprobs(v,frag1,frag2,child);
            child = child->next;
        }
        
    }
    else { /* leaf */
        ns = ((treest*)(tree->data))->index;
        q = ((treest*)(tree->data))->depth;
        frag = NULL;
        if (frag1->seq[ns]) 
            frag = frag1;
        else if (frag2 && frag2->seq[ns]) 
            frag = frag2;
        if (frag) {
            for (n=0; n<frag->seqlen; n++) {
                m = base2num(frag->seq[ns][n]);
                if (m<0) {
                    for (anc=0; anc<4; anc++)
                        frag->transprob[ns][n*4+anc] = 0.25;
                    continue;
                }
                if (n==0)
                    mutprob(v,'\0',frag->seq[ns][n+1],mutmat);
                else if (n==(frag->seqlen-1))
                    mutprob(v,frag->seq[ns][n-1],'\0',mutmat);
                else
                    mutprob(v,frag->seq[ns][n-1],frag->seq[ns][n+1],mutmat);
                transitionprob(mutmat,q,transmat);
                for (anc=0; anc<4; anc++)  {
                    frag->transprob[ns][n*4+anc] = transmat[m*4+anc];
                    if (transmat[m*4+anc]==0)
                        printf("problem!\n");
                }
            }
        }   
    }
}






void set_likelihoods(params *v, seq_frag **frag) 
{
    int n;
    GNode *pSubTree = NULL, *pNode = NULL;
    
    getClosestAncestor(v, &pNode, (*frag)->seqindex, (*frag)->nseqs);
    getSubTree(v, &pSubTree, (*frag)->seqindex, (*frag)->nseqs, &pNode);
    settransprobs(v, (*frag), NULL, pSubTree);
    
    for (n=0; n<(*frag)->seqlen; n++) 
        (*frag)->likelihood[n] = getTreeProb(v,pSubTree,(*frag),NULL,n,-1);
   
}


